; Timezone.asm
;
; Assuming this Pilot is in the UK we work out and display the time both
; here and in a few other cities around the world.
;
; This code represents a (fairly clumsy) first attempt at programming the
; USR Pilot using the truely excellent 'pila/pilrc' and the ASDK suite.
;
; It isn't very pretty and it could be tidied up and improved no end and
; also borrows heavily from the "sample" hex converter program shipped with
; the ASDK for which I thank Darrin Massena
;
; It does however work as far as I can tell and versions 1.0 onwards are
; stable.
;
; Formatted for 8-space tabs. Assemble with "pila timezone.asm" after creating
; the resources with "pilrc timezone.rcp ."
;
; By Nick Wright ("Fluffy Monster" Enterprises Ltd) from a design by Darrin
;                                                                    Massena
; 21 Nov 97 - Original Design
; 26 Nov 97 - Ver 1.0     Released
; 26 Nov 97 - Ver 1.1     Add hardcoded daylight saving features and tidy up
; 28 Nov 97 - Ver 1.1a    Small fix to the label numbers in the .RCP file
; 04 Dec 97 - Ver 1.2 i)  Moved some time conversion code into additional
;                         routine
;		      ii) Massive tidying up of the Summer/Winter code
;                     iii)Amended README to indicate which cities are covered
;                     iv) Tried to prevent crash on PalmOS 1.x
;                     v)  Added Control Event after initial Form Load to
;                         prevent initial delay in updating screen.
;                     vi) Saved so much space in the tidy-up that I decided
;                         to add New York as an example of a negative offset

; The 'Appl' directive sets the application's name and four character id.

        Appl    "Timezone", 'timz'

; Pilot.inc contains PalmOS constants, structure offsets, and API trap codes.

        include "Pilot.inc"

; Timezone.inc contains some application specific structures.

        include "Timezone.inc"

; Startup.inc contains a standard startup function. This function must be
; the first within an application and is called by the PalmOS after the app
; is loaded. The startup function in Startup.inc (__Startup__) calls the
; application-defined function PilotMain.

        include "Startup.inc"


; Application-defined resource ids

kidbAIB         equ             $7FFE
kidrPREF        equ             1
kidrTVER        equ             1
kidrTAIB        equ             1000
kidfMain        equ             1000
kidmMain        equ             1000

kidmAbout       equ             1002
kidcTimeNow     equ             1003

kidcTimeLDN     equ             1004
kidcTimeTKY     equ             1005
kidcTimeHKG     equ             1006
kidcTimeSYD     equ             1007
kidcTimeNYC     equ             1008

kidrAboutAlert  equ             1000
kidrDebugErrorAlert equ         1001

; Application-defined constants

timzDisplayLength       equ     17
timzWinterOffsetTKY     equ     32400
timzSummerOffsetTKY     equ     28800
timzWinterOffsetHKG     equ     28800
timzSummerOffsetHKG     equ     25200
timzWinterOffsetSYD     equ     39600
timzSummerOffsetSYD     equ     32400
timzWinterOffsetNYC     equ     -18000
timzSummerOffsetNYC     equ     -18000
timzZero                equ     $30
timzSpace               equ     $20
timzColon               equ     $3a
timzSlash               equ     $2f

; Some global variables (a5-relative)

        data

global dumevt.EventType
global gpfldLondon.l                    ;pointer to the London field
global gpfldTokyo.l                     ;pointer to the Tokyo field
global gpfldHongK.l                     ;pointer to the HK field
global gpfldSydney.l                    ;pointer to the Sydney field
global gpfldNewYork.l                   ;pointer to the NY field

        code

; ---------------------------------------------------------------------------
; DWord PilotMain(Word cmd, void *cmdPBP, Word launchflags)
; PilotMain is called by the startup code and implements a simple event
; handling loop.

proc PilotMain(cmd.w, cmdPBP.l, launchFlags.w)
local   err.w
local	evt.EventType

beginproc
        tst.w   cmd(a6)                 ;sysAppLaunchCmdNormalLaunch is 0
        bne     PmReturn                ;not a normal launch, bag out

        systrap FrmGotoForm(#kidfMain.w)

PmEventLoop
; Doze until an event arrives
; Start with a nilEvent which means nothing happened

        move.w  #nilEvent,evt+EventType.eType(a6)
        systrap EvtGetEvent(&evt(a6), #200.l)   ; Wait 200 clock ticks
                                                ; This is around 2 secs
        cmpi.w  #nilEvent,evt+EventType.eType(a6)
        bne.s   PmEventHandle                   ; Something happened so deal
                                                ; with it

        ; Nothing happened for 200 ticks so time to force our own event
        ; and update our display clocks. This use of EvtGetEvent with a
        ; timeout is suggested for animation use (and an updating clock
        ; is simple animation after all !). The event we force is a
        ; ctlSelectEvent which is usually used to indicate a control
        ; mechanism of some sort such as pressing a button. We just use
        ; it as a way of forcing something to happen - see later.

        move.w  #ctlSelectEvent,dumevt+EventType.eType(a5)
        systrap EvtAddEventToQueue(&dumevt(a5))
        bra.s   PmEventLoop                     ; And round again
        
PmEventHandle
; System gets first chance to handle the event

        systrap SysHandleEvent(&evt(a6))

        tst.b   d0                      ;handled?
        bne.s   PmEventDone             ;yep

; Menu handler gets second chance to handle the event

        systrap MenuHandleEvent(&0, &evt(a6), &err(a6))

        tst.b   d0                      ;handled?
        bne.s   PmEventDone             ;yep
        bra.s   PmThirdChance

PmCheat ; This is here to get round a maximum range on a branch
        bra.s   PmEventLoop

PmThirdChance
; Application handler gets third chance to handle the event

        call    ApplicationHandleEvent(&evt(a6))

        tst.b   d0                      ;handled?
        bne.s   PmEventDone             ;yep

; Form handler gets fourth chance to handle the event

        call    MainFormHandleEvent(&evt(a6))

        tst.b   d0                      ;handled?
        bne.s   PmEventDone             ;yep

; Still not handled. We're not interested in it anymore so let the
; default form handler take it.

        systrap FrmGetActiveForm()
        systrap FrmHandleEvent(a0.l, &evt(a6))

; Return from PilotMain when an appStopEvent is received

PmEventDone
        cmpi.w  #appStopEvent,evt+EventType.eType(a6) ;time to stop?
        bne.s   PmCheat                               ;no - jump back

PmReturn
        moveq   #0,d0
endproc


; ---------------------------------------------------------------------------
; Boolean ApplicationHandleEvent(EventType *pevt)
; Handles these events:
;       frmLoadEvent

proc ApplicationHandleEvent(pevt.l)
beginproc
        movem.l a3,-(a7)                ;save registers we're going to trash

        movea.l pevt(a6),a0             ;a0 = pevt

; Handle frmLoadEvents

        cmpi.w  #frmLoadEvent,EventType.eType(a0) ;frmLoadEvent?
        bne     AHENotHandled           ;no

; Initialize the form and make it active

        systrap FrmInitForm(EventType.data+frmLoad.formID(a0).w)
        move.l  a0,a3                   ;save a copy of FormPtr for later

        systrap FrmSetActiveForm(a0.l)

; Get pointers to the fields we'll be working with (London, Tokyo, HK, Sydney)

        systrap FrmGetObjectIndex(a3.l, #kidcTimeLDN.w)
        systrap FrmGetObjectPtr(a3.l,d0.w)
        move.l  a0,gpfldLondon(a5)      ;save the field ptr away

        systrap FrmGetObjectIndex(a3.l, #kidcTimeTKY.w)
        systrap FrmGetObjectPtr(a3.l,d0.w)
        move.l  a0,gpfldTokyo(a5)       ;save the field ptr away

        systrap FrmGetObjectIndex(a3.l, #kidcTimeHKG.w)
        systrap FrmGetObjectPtr(a3.l,d0.w)
        move.l  a0,gpfldHongK(a5)       ;save the field ptr away

        systrap FrmGetObjectIndex(a3.l, #kidcTimeSYD.w)
        systrap FrmGetObjectPtr(a3.l,d0.w)
        move.l  a0,gpfldSydney(a5)      ;save the field ptr away

        systrap FrmGetObjectIndex(a3.l, #kidcTimeNYC.w)
        systrap FrmGetObjectPtr(a3.l,d0.w)
        move.l  a0,gpfldNewYork(a5)     ;save the field ptr away

        moveq.l #1,d0                   ;event handled
        bra.s   AHEReturn

AHENotHandled
        clr.b   d0

AHEReturn
        movem.l (a7)+,a3
endproc


; ---------------------------------------------------------------------------
; Boolean MainFormHandleEvent(EventType *pevt)
; Handles these events:
;

proc MainFormHandleEvent(pevt.l)
beginproc

        movea.l pevt(a6),a0             ;a0 = pevt
        move.w  EventType.eType(a0),d0

; Handle frmOpenEvent

        cmp.w   #frmOpenEvent,d0
        bne     MFH1

; Draw the form

        systrap FrmGetActiveForm()
        systrap FrmDrawForm(a0.l)

; Force initial screen update event - saves waiting 2 seconds or so at the
; start of the program
        
        move.w  #ctlSelectEvent,dumevt+EventType.eType(a5)
        systrap EvtAddEventToQueue(&dumevt(a5))

        moveq.l #1,d0
        bra     MFHReturn

; Handle frmMenuEvent
MFH1
        cmp.w   #menuEvent,d0
        bne     MFH2

        cmp.w   #kidmAbout,EventType.data+menu.itemID(a0)
        bne     MFHNotHandled

        systrap FrmAlert(#kidrAboutAlert.w)

        moveq.l #1,d0
        bra     MFHReturn

; Handle ctlSelectEvent - We don't bother working out what type of
; control event since there aren't any buttons to push in our app so
; this must be the dummy event used to update the clockfaces. See the
; original 'sample.asm' for ways of working out what type of control event
; actually happened.
MFH2
        cmp.w   #ctlSelectEvent,d0
        bne     MFHNotHandled

        jsr     GetTime(pc)             ;yes, get the time and display
        bra     MFHReturn

MFHNotHandled
        clr.b   d0

MFHReturn
endproc


; ---------------------------------------------------------------------------
; void GetTime(void)
;

proc GetTime()

; This module assumes that the Pilot is always on London time - if you
; want to be more clever then modify accordingly.

; Storage for the cities
local   ldntim.l
local   toktim.l
local   hkgtim.l
local   sydtim.l
local   nyctim.l

; Working storage
local   curtim.DateTimeType                     ; See pilot.inc for def
local   distim.TimezoneType                     ; See timezone.inc for def

beginproc
        movem.l d1-d3/a3,-(a7)                  ; Save registers

; Yes. Get the Time
        clr.l   d0
        systrap TimGetSeconds()                 ; Get time in seconds
        tst.l   d0
        bne     TIM2
        systrap FrmAlert(#kidrDebugErrorAlert.w) ; Something is bad !!!
        bra     TIMReturn

TIM2                                            ; Save the time away
        move.l  d0,ldntim(a6)
        move.l  d0,toktim(a6)
        move.l  d0,hkgtim(a6)
        move.l  d0,sydtim(a6)
        move.l  d0,nyctim(a6)

; Now we need to know which offsets to add for the cities
; Assumes daylight saving changes take place on the last Sundays in
; March and October at 0000 in the morning (midnight)
; If you want anything else you'll have to play with the code
; Convert our time from seconds to DateTime
        lea.l   curtim(a6),a0
        systrap TimSecondsToDateTime(d0.l,a0)
        cmpi.w  #october,curtim.month(a6)
        beq     TIM201                  ; October gets special consideration
        bgt     TIM21                   ; Always Winter after October
        cmpi.w  #march,curtim.month(a6)
        beq     TIM202                  ; March gets special consideration
        blt     TIM21                   ; Always Winter before March
; It's between April and September so its definitely summer
        bra     TIM22

TIM201  ; It's October but is it after the 25th which is the earliest date
        ; possible for the change
        cmpi.w   #25,curtim.day(a6)
        blt     TIM22                   ; Still Summer
        bgt     TIM2011
        cmpi.w  #sunday,curtim.weekDay(a6)
        bne     TIM22                   ; Still Summer
        beq     TIM21                   ; Now Winter
TIM2011
	clr.l	d0
	clr.l	d3
	move.w	#26,d0
	move.w	#monday,d3
TIM2012
	cmp.w	curtim.day(a6),d0
        blt     TIM2013
	cmp.w	curtim.weekDay(a6),d3
        blt     TIM22                   ; Still Summer
	bra	TIM21			; Now Winter
TIM2013
	cmpi.w	#30,curtim.day(a6)
	bgt	TIM21			; Must now be winter as its Oct31
	addi.w	#1,d0
	addi.w	#1,d3
	bra	TIM2012

TIM202  ; It's March but is it after the 25th which is the earliest date
        ; possible for the change
        cmpi.w   #25,curtim.day(a6)
        blt     TIM21                   ; Still Winter
        bgt     TIM2021
        cmpi.w   #sunday,curtim.weekDay(a6)
        bne     TIM21                   ; Still Winter
        beq     TIM22                   ; Now Summer
TIM2021
	clr.l	d0
	clr.l	d3
	move.w	#26,d0
	move.w	#monday,d3
TIM2022
	cmp.w	curtim.day(a6),d0
        blt     TIM2023
	cmp.w	curtim.weekDay(a6),d3
        blt     TIM21                   ; Still Winter 
	bra	TIM22			; Now Summer
TIM2023
	cmpi.w	#30,curtim.day(a6)
	bgt	TIM22			; Must now be summer as its Mar31
	addi.w	#1,d0
	addi.w	#1,d3
	bra	TIM2022

TIM21
; It is Winter here (London)

        addi.l  #timzWinterOffsetTKY,toktim(a6) ; Simple conversion for Tokyo
        addi.l  #timzWinterOffsetHKG,hkgtim(a6) ; Equally for Hong Kong
        addi.l  #timzWinterOffsetSYD,sydtim(a6) ; And Sydney
        addi.l  #timzWinterOffsetNYC,nyctim(a6) ; And New York -ve offset
        bra     TIM23

TIM22
; It is Summer here (London)

        addi.l  #timzSummerOffsetTKY,toktim(a6) ; Simple conversion for Tokyo
        addi.l  #timzSummerOffsetHKG,hkgtim(a6) ; Equally for Hong Kong
        addi.l  #timzSummerOffsetSYD,sydtim(a6) ; And Sydney
        addi.l  #timzSummerOffsetNYC,nyctim(a6) ; And New York -ve offset

TIM23
        move.l  ldntim(a6),d0
        tst.l   d0
        bne     TIM3
        move.l  toktim(a6),d0
        tst.l   d0
        bne     TIM3
        move.l  hkgtim(a6),d0
        tst.l   d0
        bne     TIM3
        move.l  sydtim(a6),d0
        tst.l   d0
        bne     TIM3
        move.l  nyctim(a6),d0
        tst.l   d0
        bne     TIM3
        bra     TIMReturn                       ; Done all cities this pass

TIM3
        lea.l   curtim(a6),a0
        systrap TimSecondsToDateTime(d0.l,a0)
; Convert Hours
        clr.l   d0
        call    ConvertTime(curtim.hour(a6).w, &distim.TimezoneHr(a6))

; Convert Minutes
        clr.l   d0
        call    ConvertTime(curtim.minute(a6).w, &distim.TimezoneMin(a6))

; Convert Date
        clr.l   d0
        call    ConvertTime(curtim.day(a6).w, &distim.TimezoneDay(a6))

; Convert Month
        clr.l   d0
        call    ConvertTime(curtim.month(a6).w, &distim.TimezoneMnth(a6))

TIM3year                                     ; Convert Year
        clr.l   d0
        move.w  curtim.year(a6),d0
        move.l  d0,-(a7)                ;push integer
        pea.l   distim.TimezoneYr(a6)   ;push pointer to string buffer
        trap    #15
        dc.w    sysTrapStrIToA
        addq.l  #8,a7                   ;pop the args off the stack

; Place the new ASCII strings in the appropriate field depending which city
        tst.l   ldntim(a6)
        bne     TIM3London
        tst.l   toktim(a6)
        bne     TIM3Tokyo
        tst.l   hkgtim(a6)
        bne     TIM3HK
        tst.l   sydtim(a6)
        bne     TIM3Sydney
        tst.l   nyctim(a6)
        bne     TIM3NewYork
        bra     TIMReturn               ; All done this pass

TIM3London
        move.l  gpfldLondon(a5),a3      ;London field
        clr.l   ldntim(a6)
        bra     TIM4

TIM3Tokyo
        move.l  gpfldTokyo(a5),a3       ; Tokyo field
        clr.l   toktim(a6)
        bra     TIM4

TIM3HK
        move.l  gpfldHongK(a5),a3       ; HK field
        clr.l   hkgtim(a6)
        bra     TIM4

TIM3Sydney
        move.l  gpfldSydney(a5),a3      ; Sydney field
        clr.l   sydtim(a6)
        bra     TIM4

TIM3NewYork
        move.l  gpfldNewYork(a5),a3     ; NYC Time
        clr.l   nyctim(a6)

TIM4
        ; erase the old string

        systrap FldFreeMemory(a3.l)

        ; insert the new string

        move.b  #timzColon,distim.TimezoneFill1(a6)
        move.b  #timzSpace,distim.TimezoneFill2(a6)
        move.b  #timzSpace,1+distim.TimezoneFill2(a6)
        move.b  #timzSlash,distim.TimezoneFill3(a6)
        move.b  #timzSlash,distim.TimezoneFill4(a6)
        move.w  #timzDisplayLength,d0
        lea.l   distim(a6),a0
        systrap FldInsert(a3.l,a0.l,d0.w)
        bra     TIM23

TIMReturn
        ; Clear the dirty bits on all fields

        systrap FldSetDirty(gpfldLondon(a5).l,#0.b)
        systrap FldSetDirty(gpfldTokyo(a5).l,#0.b)
        systrap FldSetDirty(gpfldHongK(a5).l,#0.b)
        systrap FldSetDirty(gpfldSydney(a5).l,#0.b)
        systrap FldSetDirty(gpfldNewYork(a5).l,#0.b)

        movem.l  (a7)+,d1-d3/a3
endproc

proc    ConvertTime(cint.w, saddr.l)

; A little conversion routine to replace a lot of code in earlier versions
; This routine does the Integer to String conversion for the two digit
; values (hour,minute,date,month)

beginproc
        move.w  cint(a6),d0             ; Retrieve the arguments
        movea.l saddr(a6),a0
        move.l  d0,-(a7)                ; Push onto stack
        move.l  a0,-(a7)
        trap    #15
        dc.w    sysTrapStrIToA          ;Call the system conversion
        addq.l  #8,a7                   ;pop the args off the stack
        tst.b   1(a0)                   ;need to add leading zero ?
        bne     ConvertTimeEnd          ;Nope
        move.b  (a0),1(a0)              ;Yep - move the first byte
        move.b  #timzZero,(a0)          ;Add the leading zero

ConvertTimeEnd

endproc
;
; Resources -----------------------------------------------------------------
;

; Application Icon Bitmap resource. Is automatically converted from Windows
; format to Pilot format and written as a 'tAIB' rather than 'Tbmp' because
; kidbAIB is a special value ($7ffe)

        res 'WBMP', kidbAIB, "Timezone.bmp"

; 'pref' resource. Defines app launch flags, stack and heap size

        res 'pref', kidrPREF
        dc.w    sysAppLaunchFlagNewStack|sysAppLaunchFlagNewGlobals|sysAppLaunchFlagUIApp|sysAppLaunchFlagSubCall
        dc.l    $1000                           ; stack size
        dc.l    $1000                           ; heap size

; Form resource

        res 'tFRM', kidfMain, "tfrm03e8.bin"

; Version resource

        res 'tver', kidrTVER
        dc.b    '1.0', 0

; Menu resource

        res 'MBAR', kidmMain, "mbar03e8.bin"

; Alert resources

        res 'Talt', kidrAboutAlert, "talt03e8.bin"
        res 'Talt', kidrDebugErrorAlert, "talt03e9.bin"
